# Copyright (c) 2016, Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
import uuid

from osc_lib import exceptions

from openstackclient.network.v2 import network_qos_rule
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


RULE_TYPE_BANDWIDTH_LIMIT = 'bandwidth-limit'
RULE_TYPE_DSCP_MARKING = 'dscp-marking'
RULE_TYPE_MINIMUM_BANDWIDTH = 'minimum-bandwidth'
RULE_TYPE_MINIMUM_PACKET_RATE = 'minimum-packet-rate'
DSCP_VALID_MARKS = [
    0,
    8,
    10,
    12,
    14,
    16,
    18,
    20,
    22,
    24,
    26,
    28,
    30,
    32,
    34,
    36,
    38,
    40,
    46,
    48,
    56,
]


class TestNetworkQosRule(network_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()
        self.qos_policy = network_fakes.create_one_qos_policy()
        self.network_client.find_qos_policy.return_value = self.qos_policy


class TestCreateNetworkQosRuleMinimumBandwidth(TestNetworkQosRule):
    def test_check_type_parameters(self):
        pass

    def setUp(self):
        super().setUp()
        attrs = {
            'qos_policy_id': self.qos_policy.id,
            'type': RULE_TYPE_MINIMUM_BANDWIDTH,
        }
        self.new_rule = network_fakes.create_one_qos_rule(attrs)
        self.columns = (
            'direction',
            'id',
            'min_kbps',
            'project_id',
            'type',
        )

        self.data = (
            self.new_rule.direction,
            self.new_rule.id,
            self.new_rule.min_kbps,
            self.new_rule.project_id,
            self.new_rule.type,
        )
        self.network_client.create_qos_minimum_bandwidth_rule.return_value = (
            self.new_rule
        )

        # Get the command object to test
        self.cmd = network_qos_rule.CreateNetworkQosRule(self.app, None)

    def test_create_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_default_options(self):
        arglist = [
            '--type',
            RULE_TYPE_MINIMUM_BANDWIDTH,
            '--min-kbps',
            str(self.new_rule.min_kbps),
            '--egress',
            self.new_rule.qos_policy_id,
        ]

        verifylist = [
            ('type', RULE_TYPE_MINIMUM_BANDWIDTH),
            ('min_kbps', self.new_rule.min_kbps),
            ('egress', True),
            ('qos_policy', self.new_rule.qos_policy_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_qos_minimum_bandwidth_rule.assert_called_once_with(
            self.qos_policy.id,
            **{
                'min_kbps': self.new_rule.min_kbps,
                'direction': self.new_rule.direction,
            },
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_wrong_options(self):
        arglist = [
            '--type',
            RULE_TYPE_MINIMUM_BANDWIDTH,
            '--max-kbps',
            '10000',
            self.new_rule.qos_policy_id,
        ]

        verifylist = [
            ('type', RULE_TYPE_MINIMUM_BANDWIDTH),
            ('max_kbps', 10000),
            ('qos_policy', self.new_rule.qos_policy_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        try:
            self.cmd.take_action(parsed_args)
        except exceptions.CommandError as e:
            msg = (
                'Failed to create Network QoS rule: "Create" rule command '
                'for type "minimum-bandwidth" requires arguments: '
                'direction, min_kbps'
            )
            self.assertEqual(msg, str(e))


class TestCreateNetworkQosRuleMinimumPacketRate(TestNetworkQosRule):
    def test_check_type_parameters(self):
        pass

    def setUp(self):
        super().setUp()
        attrs = {
            'qos_policy_id': self.qos_policy.id,
            'type': RULE_TYPE_MINIMUM_PACKET_RATE,
        }
        self.new_rule = network_fakes.create_one_qos_rule(attrs)
        self.columns = (
            'direction',
            'id',
            'min_kpps',
            'project_id',
            'type',
        )

        self.data = (
            self.new_rule.direction,
            self.new_rule.id,
            self.new_rule.min_kpps,
            self.new_rule.project_id,
            self.new_rule.type,
        )
        self.network_client.create_qos_minimum_packet_rate_rule.return_value = self.new_rule

        # Get the command object to test
        self.cmd = network_qos_rule.CreateNetworkQosRule(self.app, None)

    def test_create_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_default_options(self):
        arglist = [
            '--type',
            RULE_TYPE_MINIMUM_PACKET_RATE,
            '--min-kpps',
            str(self.new_rule.min_kpps),
            '--egress',
            self.new_rule.qos_policy_id,
        ]

        verifylist = [
            ('type', RULE_TYPE_MINIMUM_PACKET_RATE),
            ('min_kpps', self.new_rule.min_kpps),
            ('egress', True),
            ('qos_policy', self.new_rule.qos_policy_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_qos_minimum_packet_rate_rule.assert_called_once_with(
            self.qos_policy.id,
            **{
                'min_kpps': self.new_rule.min_kpps,
                'direction': self.new_rule.direction,
            },
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_wrong_options(self):
        arglist = [
            '--type',
            RULE_TYPE_MINIMUM_PACKET_RATE,
            '--min-kbps',
            '10000',
            self.new_rule.qos_policy_id,
        ]

        verifylist = [
            ('type', RULE_TYPE_MINIMUM_PACKET_RATE),
            ('min_kbps', 10000),
            ('qos_policy', self.new_rule.qos_policy_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        try:
            self.cmd.take_action(parsed_args)
        except exceptions.CommandError as e:
            msg = (
                'Failed to create Network QoS rule: "Create" rule command '
                'for type "minimum-packet-rate" requires arguments: '
                'direction, min_kpps'
            )
            self.assertEqual(msg, str(e))


class TestCreateNetworkQosRuleDSCPMarking(TestNetworkQosRule):
    def test_check_type_parameters(self):
        pass

    def setUp(self):
        super().setUp()
        attrs = {
            'qos_policy_id': self.qos_policy.id,
            'type': RULE_TYPE_DSCP_MARKING,
        }
        self.new_rule = network_fakes.create_one_qos_rule(attrs)
        self.columns = (
            'dscp_mark',
            'id',
            'project_id',
            'type',
        )

        self.data = (
            self.new_rule.dscp_mark,
            self.new_rule.id,
            self.new_rule.project_id,
            self.new_rule.type,
        )
        self.network_client.create_qos_dscp_marking_rule.return_value = (
            self.new_rule
        )

        # Get the command object to test
        self.cmd = network_qos_rule.CreateNetworkQosRule(self.app, None)

    def test_create_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_default_options(self):
        arglist = [
            '--type',
            RULE_TYPE_DSCP_MARKING,
            '--dscp-mark',
            str(self.new_rule.dscp_mark),
            self.new_rule.qos_policy_id,
        ]

        verifylist = [
            ('type', RULE_TYPE_DSCP_MARKING),
            ('dscp_mark', self.new_rule.dscp_mark),
            ('qos_policy', self.new_rule.qos_policy_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_qos_dscp_marking_rule.assert_called_once_with(
            self.qos_policy.id, **{'dscp_mark': self.new_rule.dscp_mark}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_wrong_options(self):
        arglist = [
            '--type',
            RULE_TYPE_DSCP_MARKING,
            '--max-kbps',
            '10000',
            self.new_rule.qos_policy_id,
        ]

        verifylist = [
            ('type', RULE_TYPE_DSCP_MARKING),
            ('max_kbps', 10000),
            ('qos_policy', self.new_rule.qos_policy_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        try:
            self.cmd.take_action(parsed_args)
        except exceptions.CommandError as e:
            msg = (
                'Failed to create Network QoS rule: "Create" rule command '
                'for type "dscp-marking" requires arguments: dscp_mark'
            )
            self.assertEqual(msg, str(e))


class TestCreateNetworkQosRuleBandwidtLimit(TestNetworkQosRule):
    def test_check_type_parameters(self):
        pass

    def setUp(self):
        super().setUp()
        attrs = {
            'qos_policy_id': self.qos_policy.id,
            'type': RULE_TYPE_BANDWIDTH_LIMIT,
        }
        self.new_rule = network_fakes.create_one_qos_rule(attrs)
        self.columns = (
            'direction',
            'id',
            'max_burst_kbps',
            'max_kbps',
            'project_id',
            'type',
        )

        self.data = (
            self.new_rule.direction,
            self.new_rule.id,
            self.new_rule.max_burst_kbps,
            self.new_rule.max_kbps,
            self.new_rule.project_id,
            self.new_rule.type,
        )
        self.network_client.create_qos_bandwidth_limit_rule.return_value = (
            self.new_rule
        )

        # Get the command object to test
        self.cmd = network_qos_rule.CreateNetworkQosRule(self.app, None)

    def test_create_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_default_options(self):
        arglist = [
            '--type',
            RULE_TYPE_BANDWIDTH_LIMIT,
            '--max-kbps',
            str(self.new_rule.max_kbps),
            '--egress',
            self.new_rule.qos_policy_id,
        ]

        verifylist = [
            ('type', RULE_TYPE_BANDWIDTH_LIMIT),
            ('max_kbps', self.new_rule.max_kbps),
            ('egress', True),
            ('qos_policy', self.new_rule.qos_policy_id),
        ]

        rule = network_fakes.create_one_qos_rule(
            {
                'qos_policy_id': self.qos_policy.id,
                'type': RULE_TYPE_BANDWIDTH_LIMIT,
            }
        )
        rule.max_burst_kbps = 0
        expected_data = (
            rule.direction,
            rule.id,
            rule.max_burst_kbps,
            rule.max_kbps,
            rule.project_id,
            rule.type,
        )

        with mock.patch.object(
            self.network_client,
            "create_qos_bandwidth_limit_rule",
            return_value=rule,
        ) as create_qos_bandwidth_limit_rule:
            parsed_args = self.check_parser(self.cmd, arglist, verifylist)
            columns, data = self.cmd.take_action(parsed_args)

        create_qos_bandwidth_limit_rule.assert_called_once_with(
            self.qos_policy.id,
            **{
                'max_kbps': self.new_rule.max_kbps,
                'direction': self.new_rule.direction,
            },
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(expected_data, data)

    def test_create_all_options(self):
        arglist = [
            '--type',
            RULE_TYPE_BANDWIDTH_LIMIT,
            '--max-kbps',
            str(self.new_rule.max_kbps),
            '--max-burst-kbits',
            str(self.new_rule.max_burst_kbps),
            '--egress',
            self.new_rule.qos_policy_id,
        ]

        verifylist = [
            ('type', RULE_TYPE_BANDWIDTH_LIMIT),
            ('max_kbps', self.new_rule.max_kbps),
            ('max_burst_kbits', self.new_rule.max_burst_kbps),
            ('egress', True),
            ('qos_policy', self.new_rule.qos_policy_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_qos_bandwidth_limit_rule.assert_called_once_with(
            self.qos_policy.id,
            **{
                'max_kbps': self.new_rule.max_kbps,
                'max_burst_kbps': self.new_rule.max_burst_kbps,
                'direction': self.new_rule.direction,
            },
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_wrong_options(self):
        arglist = [
            '--type',
            RULE_TYPE_BANDWIDTH_LIMIT,
            '--min-kbps',
            '10000',
            self.new_rule.qos_policy_id,
        ]

        verifylist = [
            ('type', RULE_TYPE_BANDWIDTH_LIMIT),
            ('min_kbps', 10000),
            ('qos_policy', self.new_rule.qos_policy_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        try:
            self.cmd.take_action(parsed_args)
        except exceptions.CommandError as e:
            msg = (
                'Failed to create Network QoS rule: "Create" rule command '
                'for type "bandwidth-limit" requires arguments: max_kbps'
            )
            self.assertEqual(msg, str(e))


class TestDeleteNetworkQosRuleMinimumBandwidth(TestNetworkQosRule):
    def setUp(self):
        super().setUp()
        attrs = {
            'qos_policy_id': self.qos_policy.id,
            'type': RULE_TYPE_MINIMUM_BANDWIDTH,
        }
        self.new_rule = network_fakes.create_one_qos_rule(attrs)
        self.qos_policy.rules = [self.new_rule]
        self.network_client.delete_qos_minimum_bandwidth_rule.return_value = (
            None
        )

        self.network_client.find_qos_minimum_bandwidth_rule = (
            network_fakes.get_qos_rules(qos_rules=self.new_rule)
        )

        # Get the command object to test
        self.cmd = network_qos_rule.DeleteNetworkQosRule(self.app, None)

    def test_qos_policy_delete(self):
        arglist = [
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.find_qos_policy.assert_called_once_with(
            self.qos_policy.id, ignore_missing=False
        )
        self.network_client.delete_qos_minimum_bandwidth_rule.assert_called_once_with(
            self.new_rule.id, self.qos_policy.id
        )
        self.assertIsNone(result)

    def test_qos_policy_delete_error(self):
        arglist = [
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        self.network_client.delete_qos_minimum_bandwidth_rule.side_effect = (
            Exception('Error message')
        )
        try:
            parsed_args = self.check_parser(self.cmd, arglist, verifylist)
            self.cmd.take_action(parsed_args)
        except exceptions.CommandError as e:
            msg = 'Failed to delete Network QoS rule ID "{rule}": {e}'.format(
                rule=self.new_rule.id,
                e='Error message',
            )
            self.assertEqual(msg, str(e))


class TestDeleteNetworkQosRuleMinimumPacketRate(TestNetworkQosRule):
    def setUp(self):
        super().setUp()
        attrs = {
            'qos_policy_id': self.qos_policy.id,
            'type': RULE_TYPE_MINIMUM_PACKET_RATE,
        }
        self.new_rule = network_fakes.create_one_qos_rule(attrs)
        self.qos_policy.rules = [self.new_rule]
        self.network_client.delete_qos_minimum_packet_rate_rule.return_value = None

        self.network_client.find_qos_minimum_packet_rate_rule = (
            network_fakes.get_qos_rules(qos_rules=self.new_rule)
        )

        # Get the command object to test
        self.cmd = network_qos_rule.DeleteNetworkQosRule(self.app, None)

    def test_qos_policy_delete(self):
        arglist = [
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.find_qos_policy.assert_called_once_with(
            self.qos_policy.id, ignore_missing=False
        )
        self.network_client.delete_qos_minimum_packet_rate_rule.assert_called_once_with(
            self.new_rule.id, self.qos_policy.id
        )
        self.assertIsNone(result)

    def test_qos_policy_delete_error(self):
        arglist = [
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        self.network_client.delete_qos_minimum_packet_rate_rule.side_effect = (
            Exception('Error message')
        )
        try:
            parsed_args = self.check_parser(self.cmd, arglist, verifylist)
            self.cmd.take_action(parsed_args)
        except exceptions.CommandError as e:
            msg = 'Failed to delete Network QoS rule ID "{rule}": {e}'.format(
                rule=self.new_rule.id,
                e='Error message',
            )
            self.assertEqual(msg, str(e))


class TestDeleteNetworkQosRuleDSCPMarking(TestNetworkQosRule):
    def setUp(self):
        super().setUp()
        attrs = {
            'qos_policy_id': self.qos_policy.id,
            'type': RULE_TYPE_DSCP_MARKING,
        }
        self.new_rule = network_fakes.create_one_qos_rule(attrs)
        self.qos_policy.rules = [self.new_rule]
        self.network_client.delete_qos_dscp_marking_rule.return_value = None

        self.network_client.find_qos_dscp_marking_rule = (
            network_fakes.get_qos_rules(qos_rules=self.new_rule)
        )

        # Get the command object to test
        self.cmd = network_qos_rule.DeleteNetworkQosRule(self.app, None)

    def test_qos_policy_delete(self):
        arglist = [
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.find_qos_policy.assert_called_once_with(
            self.qos_policy.id, ignore_missing=False
        )
        self.network_client.delete_qos_dscp_marking_rule.assert_called_once_with(
            self.new_rule.id, self.qos_policy.id
        )
        self.assertIsNone(result)

    def test_qos_policy_delete_error(self):
        arglist = [
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        self.network_client.delete_qos_dscp_marking_rule.side_effect = (
            Exception('Error message')
        )
        try:
            parsed_args = self.check_parser(self.cmd, arglist, verifylist)
            self.cmd.take_action(parsed_args)
        except exceptions.CommandError as e:
            msg = 'Failed to delete Network QoS rule ID "{rule}": {e}'.format(
                rule=self.new_rule.id,
                e='Error message',
            )
            self.assertEqual(msg, str(e))


class TestDeleteNetworkQosRuleBandwidthLimit(TestNetworkQosRule):
    def setUp(self):
        super().setUp()
        attrs = {
            'qos_policy_id': self.qos_policy.id,
            'type': RULE_TYPE_BANDWIDTH_LIMIT,
        }
        self.new_rule = network_fakes.create_one_qos_rule(attrs)
        self.qos_policy.rules = [self.new_rule]
        self.network_client.delete_qos_bandwidth_limit_rule.return_value = None

        self.network_client.find_qos_bandwidth_limit_rule = (
            network_fakes.get_qos_rules(qos_rules=self.new_rule)
        )

        # Get the command object to test
        self.cmd = network_qos_rule.DeleteNetworkQosRule(self.app, None)

    def test_qos_policy_delete(self):
        arglist = [
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.find_qos_policy.assert_called_once_with(
            self.qos_policy.id, ignore_missing=False
        )
        self.network_client.delete_qos_bandwidth_limit_rule.assert_called_once_with(
            self.new_rule.id, self.qos_policy.id
        )
        self.assertIsNone(result)

    def test_qos_policy_delete_error(self):
        arglist = [
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        self.network_client.delete_qos_bandwidth_limit_rule.side_effect = (
            Exception('Error message')
        )
        try:
            parsed_args = self.check_parser(self.cmd, arglist, verifylist)
            self.cmd.take_action(parsed_args)
        except exceptions.CommandError as e:
            msg = 'Failed to delete Network QoS rule ID "{rule}": {e}'.format(
                rule=self.new_rule.id,
                e='Error message',
            )
            self.assertEqual(msg, str(e))


class TestSetNetworkQosRuleMinimumBandwidth(TestNetworkQosRule):
    def setUp(self):
        super().setUp()
        attrs = {
            'qos_policy_id': self.qos_policy.id,
            'type': RULE_TYPE_MINIMUM_BANDWIDTH,
        }
        self.new_rule = network_fakes.create_one_qos_rule(attrs)
        self.qos_policy.rules = [self.new_rule]
        self.network_client.update_qos_minimum_bandwidth_rule.return_value = (
            None
        )

        self.network_client.find_qos_minimum_bandwidth_rule.return_value = (
            self.new_rule
        )

        self.network_client.find_qos_policy.return_value = self.qos_policy

        # Get the command object to test
        self.cmd = network_qos_rule.SetNetworkQosRule(self.app, None)

    def test_set_nothing(self):
        arglist = [
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_qos_minimum_bandwidth_rule.assert_called_with(
            self.new_rule, self.qos_policy.id
        )
        self.assertIsNone(result)

    def test_set_min_kbps(self):
        self._set_min_kbps()

    def test_set_min_kbps_to_zero(self):
        self._set_min_kbps(min_kbps=0)

    def _set_min_kbps(self, min_kbps=None):
        if min_kbps:
            previous_min_kbps = self.new_rule.min_kbps
            self.new_rule.min_kbps = min_kbps

        arglist = [
            '--min-kbps',
            str(self.new_rule.min_kbps),
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('min_kbps', self.new_rule.min_kbps),
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {
            'min_kbps': self.new_rule.min_kbps,
        }
        self.network_client.update_qos_minimum_bandwidth_rule.assert_called_with(
            self.new_rule, self.qos_policy.id, **attrs
        )
        self.assertIsNone(result)

        if min_kbps:
            self.new_rule.min_kbps = previous_min_kbps

    def test_set_wrong_options(self):
        arglist = [
            '--max-kbps',
            str(10000),
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('max_kbps', 10000),
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        try:
            self.cmd.take_action(parsed_args)
        except exceptions.CommandError as e:
            msg = (
                f'Failed to set Network QoS rule ID "{self.new_rule.id}": Rule type '
                '"minimum-bandwidth" only requires arguments: direction, '
                'min_kbps'
            )
            self.assertEqual(msg, str(e))


class TestSetNetworkQosRuleMinimumPacketRate(TestNetworkQosRule):
    def setUp(self):
        super().setUp()
        attrs = {
            'qos_policy_id': self.qos_policy.id,
            'type': RULE_TYPE_MINIMUM_PACKET_RATE,
        }
        self.new_rule = network_fakes.create_one_qos_rule(attrs)
        self.qos_policy.rules = [self.new_rule]
        self.network_client.update_qos_minimum_packet_rate_rule.return_value = None

        self.network_client.find_qos_minimum_packet_rate_rule.return_value = (
            self.new_rule
        )

        self.network_client.find_qos_policy.return_value = self.qos_policy

        # Get the command object to test
        self.cmd = network_qos_rule.SetNetworkQosRule(self.app, None)

    def test_set_nothing(self):
        arglist = [
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_qos_minimum_packet_rate_rule.assert_called_with(
            self.new_rule, self.qos_policy.id
        )
        self.assertIsNone(result)

    def test_set_min_kpps(self):
        self._set_min_kpps()

    def test_set_min_kpps_to_zero(self):
        self._set_min_kpps(min_kpps=0)

    def _set_min_kpps(self, min_kpps=None):
        if min_kpps:
            previous_min_kpps = self.new_rule.min_kpps
            self.new_rule.min_kpps = min_kpps

        arglist = [
            '--min-kpps',
            str(self.new_rule.min_kpps),
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('min_kpps', self.new_rule.min_kpps),
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {
            'min_kpps': self.new_rule.min_kpps,
        }
        self.network_client.update_qos_minimum_packet_rate_rule.assert_called_with(
            self.new_rule, self.qos_policy.id, **attrs
        )
        self.assertIsNone(result)

        if min_kpps:
            self.new_rule.min_kpps = previous_min_kpps

    def test_set_wrong_options(self):
        arglist = [
            '--min-kbps',
            str(10000),
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('min_kbps', 10000),
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        try:
            self.cmd.take_action(parsed_args)
        except exceptions.CommandError as e:
            msg = (
                f'Failed to set Network QoS rule ID "{self.new_rule.id}": Rule type '
                '"minimum-packet-rate" only requires arguments: direction, '
                'min_kpps'
            )
            self.assertEqual(msg, str(e))


class TestSetNetworkQosRuleDSCPMarking(TestNetworkQosRule):
    def setUp(self):
        super().setUp()
        attrs = {
            'qos_policy_id': self.qos_policy.id,
            'type': RULE_TYPE_DSCP_MARKING,
        }
        self.new_rule = network_fakes.create_one_qos_rule(attrs)
        self.qos_policy.rules = [self.new_rule]
        self.network_client.update_qos_dscp_marking_rule.return_value = None

        self.network_client.find_qos_dscp_marking_rule.return_value = (
            self.new_rule
        )

        self.network_client.find_qos_policy.return_value = self.qos_policy

        # Get the command object to test
        self.cmd = network_qos_rule.SetNetworkQosRule(self.app, None)

    def test_set_nothing(self):
        arglist = [
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_qos_dscp_marking_rule.assert_called_with(
            self.new_rule, self.qos_policy.id
        )
        self.assertIsNone(result)

    def test_set_dscp_mark(self):
        self._set_dscp_mark()

    def test_set_dscp_mark_to_zero(self):
        self._set_dscp_mark(dscp_mark=0)

    def _set_dscp_mark(self, dscp_mark=None):
        if dscp_mark:
            previous_dscp_mark = self.new_rule.dscp_mark
            self.new_rule.dscp_mark = dscp_mark

        arglist = [
            '--dscp-mark',
            str(self.new_rule.dscp_mark),
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('dscp_mark', self.new_rule.dscp_mark),
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {
            'dscp_mark': self.new_rule.dscp_mark,
        }
        self.network_client.update_qos_dscp_marking_rule.assert_called_with(
            self.new_rule, self.qos_policy.id, **attrs
        )
        self.assertIsNone(result)

        if dscp_mark:
            self.new_rule.dscp_mark = previous_dscp_mark

    def test_set_wrong_options(self):
        arglist = [
            '--max-kbps',
            str(10000),
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('max_kbps', 10000),
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        try:
            self.cmd.take_action(parsed_args)
        except exceptions.CommandError as e:
            msg = (
                f'Failed to set Network QoS rule ID "{self.new_rule.id}": Rule type '
                '"dscp-marking" only requires arguments: dscp_mark'
            )
            self.assertEqual(msg, str(e))


class TestSetNetworkQosRuleBandwidthLimit(TestNetworkQosRule):
    def setUp(self):
        super().setUp()
        attrs = {
            'qos_policy_id': self.qos_policy.id,
            'type': RULE_TYPE_BANDWIDTH_LIMIT,
        }
        self.new_rule = network_fakes.create_one_qos_rule(attrs)
        self.qos_policy.rules = [self.new_rule]
        self.network_client.update_qos_bandwidth_limit_rule.return_value = None
        self.network_client.find_qos_bandwidth_limit_rule.return_value = (
            self.new_rule
        )
        self.network_client.find_qos_policy.return_value = self.qos_policy

        self.cmd = network_qos_rule.SetNetworkQosRule(self.app, None)

    def test_set_nothing(self):
        arglist = [
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_qos_bandwidth_limit_rule.assert_called_with(
            self.new_rule, self.qos_policy.id
        )
        self.assertIsNone(result)

    def test_set_max_kbps(self):
        self._set_max_kbps()

    def test_set_max_kbps_to_zero(self):
        self._set_max_kbps(max_kbps=0)

    def _reset_max_kbps(self, max_kbps):
        self.new_rule.max_kbps = max_kbps

    def _set_max_kbps(self, max_kbps=None):
        if max_kbps:
            self.addCleanup(self._reset_max_kbps, self.new_rule.max_kbps)
            self.new_rule.max_kbps = max_kbps

        arglist = [
            '--max-kbps',
            str(self.new_rule.max_kbps),
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('max_kbps', self.new_rule.max_kbps),
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {
            'max_kbps': self.new_rule.max_kbps,
        }
        self.network_client.update_qos_bandwidth_limit_rule.assert_called_with(
            self.new_rule, self.qos_policy.id, **attrs
        )
        self.assertIsNone(result)

    def test_set_max_burst_kbits(self):
        self._set_max_burst_kbits()

    def test_set_max_burst_kbits_to_zero(self):
        self._set_max_burst_kbits(max_burst_kbits=0)

    def _reset_max_burst_kbits(self, max_burst_kbits):
        self.new_rule.max_burst_kbps = max_burst_kbits

    def _set_max_burst_kbits(self, max_burst_kbits=None):
        if max_burst_kbits:
            self.addCleanup(
                self._reset_max_burst_kbits, self.new_rule.max_burst_kbps
            )
            self.new_rule.max_burst_kbps = max_burst_kbits

        arglist = [
            '--max-burst-kbits',
            str(self.new_rule.max_burst_kbps),
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('max_burst_kbits', self.new_rule.max_burst_kbps),
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {
            'max_burst_kbps': self.new_rule.max_burst_kbps,
        }
        self.network_client.update_qos_bandwidth_limit_rule.assert_called_with(
            self.new_rule, self.qos_policy.id, **attrs
        )
        self.assertIsNone(result)

    def test_set_direction_egress(self):
        self._set_direction('egress')

    def test_set_direction_ingress(self):
        self._set_direction('ingress')

    def _reset_direction(self, direction):
        self.new_rule.direction = direction

    def _set_direction(self, direction):
        self.addCleanup(self._reset_direction, self.new_rule.direction)

        arglist = [
            f'--{direction}',
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            (direction, True),
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {
            'direction': direction,
        }
        self.network_client.update_qos_bandwidth_limit_rule.assert_called_with(
            self.new_rule, self.qos_policy.id, **attrs
        )
        self.assertIsNone(result)

    def test_set_wrong_options(self):
        arglist = [
            '--min-kbps',
            str(10000),
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('min_kbps', 10000),
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        try:
            self.cmd.take_action(parsed_args)
        except exceptions.CommandError as e:
            msg = (
                f'Failed to set Network QoS rule ID "{self.new_rule.id}": Rule type '
                '"bandwidth-limit" only requires arguments: direction, '
                'max_burst_kbps, max_kbps'
            )
            self.assertEqual(msg, str(e))


class TestListNetworkQosRule(TestNetworkQosRule):
    def setUp(self):
        super().setUp()
        self.qos_policy.rules = [
            {
                'max_kbps': 1024,
                'max_burst_kbps': 1024,
                'direction': 'egress',
                'id': 'qos-rule-id-' + uuid.uuid4().hex,
                'qos_policy_id': self.qos_policy.id,
                'type': 'bandwidth_limit',
            },
            {
                'dscp_mark': 0,
                'id': 'qos-rule-id-' + uuid.uuid4().hex,
                'qos_policy_id': self.qos_policy.id,
                'type': 'dscp_marking',
            },
            {
                'min_kbps': 1024,
                'direction': 'egress',
                'id': 'qos-rule-id-' + uuid.uuid4().hex,
                'qos_policy_id': self.qos_policy.id,
                'type': 'minimum_bandwidth',
            },
            {
                'min_kpps': 2800,
                'direction': 'egress',
                'id': 'qos-rule-id-' + uuid.uuid4().hex,
                'qos_policy_id': self.qos_policy.id,
                'type': 'minimum_packet_rate',
            },
        ]
        self.columns = (
            'ID',
            'QoS Policy ID',
            'Type',
            'Max Kbps',
            'Max Burst Kbits',
            'Min Kbps',
            'Min Kpps',
            'DSCP mark',
            'Direction',
        )
        self.data = []
        for index in range(len(self.qos_policy.rules)):
            self.data.append(
                (
                    self.qos_policy.rules[index]['id'],
                    self.qos_policy.id,
                    self.qos_policy.rules[index]['type'],
                    self.qos_policy.rules[index].get('max_kbps', ''),
                    self.qos_policy.rules[index].get('max_burst_kbps', ''),
                    self.qos_policy.rules[index].get('min_kbps', ''),
                    self.qos_policy.rules[index].get('min_kpps', ''),
                    self.qos_policy.rules[index].get('dscp_mark', ''),
                    self.qos_policy.rules[index].get('direction', ''),
                )
            )
        self.cmd = network_qos_rule.ListNetworkQosRule(self.app, None)

    def test_qos_rule_list(self):
        arglist = [self.qos_policy.id]
        verifylist = [
            ('qos_policy', self.qos_policy.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.find_qos_policy.assert_called_once_with(
            self.qos_policy.id, ignore_missing=False
        )
        self.assertEqual(self.columns, columns)
        list_data = list(data)
        self.assertEqual(len(self.data), len(list_data))
        for index in range(len(list_data)):
            self.assertEqual(self.data[index], list_data[index])


class TestShowNetworkQosRuleMinimumBandwidth(TestNetworkQosRule):
    def setUp(self):
        super().setUp()
        attrs = {
            'qos_policy_id': self.qos_policy.id,
            'type': RULE_TYPE_MINIMUM_BANDWIDTH,
        }
        self.new_rule = network_fakes.create_one_qos_rule(attrs)
        self.qos_policy.rules = [self.new_rule]
        self.columns = (
            'direction',
            'id',
            'min_kbps',
            'project_id',
            'type',
        )
        self.data = (
            self.new_rule.direction,
            self.new_rule.id,
            self.new_rule.min_kbps,
            self.new_rule.project_id,
            self.new_rule.type,
        )

        self.network_client.get_qos_minimum_bandwidth_rule.return_value = (
            self.new_rule
        )

        # Get the command object to test
        self.cmd = network_qos_rule.ShowNetworkQosRule(self.app, None)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_show_all_options(self):
        arglist = [
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.get_qos_minimum_bandwidth_rule.assert_called_once_with(
            self.new_rule.id, self.qos_policy.id
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.data), list(data))


class TestShowNetworkQosRuleMinimumPacketRate(TestNetworkQosRule):
    def setUp(self):
        super().setUp()
        attrs = {
            'qos_policy_id': self.qos_policy.id,
            'type': RULE_TYPE_MINIMUM_PACKET_RATE,
        }
        self.new_rule = network_fakes.create_one_qos_rule(attrs)
        self.qos_policy.rules = [self.new_rule]
        self.columns = (
            'direction',
            'id',
            'min_kpps',
            'project_id',
            'type',
        )
        self.data = (
            self.new_rule.direction,
            self.new_rule.id,
            self.new_rule.min_kpps,
            self.new_rule.project_id,
            self.new_rule.type,
        )

        self.network_client.get_qos_minimum_packet_rate_rule.return_value = (
            self.new_rule
        )

        # Get the command object to test
        self.cmd = network_qos_rule.ShowNetworkQosRule(self.app, None)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_show_all_options(self):
        arglist = [
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.get_qos_minimum_packet_rate_rule.assert_called_once_with(
            self.new_rule.id, self.qos_policy.id
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.data), list(data))


class TestShowNetworkQosDSCPMarking(TestNetworkQosRule):
    def setUp(self):
        super().setUp()
        attrs = {
            'qos_policy_id': self.qos_policy.id,
            'type': RULE_TYPE_DSCP_MARKING,
        }
        self.new_rule = network_fakes.create_one_qos_rule(attrs)
        self.qos_policy.rules = [self.new_rule]
        self.columns = (
            'dscp_mark',
            'id',
            'project_id',
            'type',
        )
        self.data = (
            self.new_rule.dscp_mark,
            self.new_rule.id,
            self.new_rule.project_id,
            self.new_rule.type,
        )

        self.network_client.get_qos_dscp_marking_rule.return_value = (
            self.new_rule
        )

        # Get the command object to test
        self.cmd = network_qos_rule.ShowNetworkQosRule(self.app, None)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_show_all_options(self):
        arglist = [
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.get_qos_dscp_marking_rule.assert_called_once_with(
            self.new_rule.id, self.qos_policy.id
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.data), list(data))


class TestShowNetworkQosBandwidthLimit(TestNetworkQosRule):
    def setUp(self):
        super().setUp()
        attrs = {
            'qos_policy_id': self.qos_policy.id,
            'type': RULE_TYPE_BANDWIDTH_LIMIT,
        }
        self.new_rule = network_fakes.create_one_qos_rule(attrs)
        self.qos_policy.rules = [self.new_rule]
        self.columns = (
            'direction',
            'id',
            'max_burst_kbps',
            'max_kbps',
            'project_id',
            'type',
        )
        self.data = (
            self.new_rule.direction,
            self.new_rule.id,
            self.new_rule.max_burst_kbps,
            self.new_rule.max_kbps,
            self.new_rule.project_id,
            self.new_rule.type,
        )

        self.network_client.get_qos_bandwidth_limit_rule.return_value = (
            self.new_rule
        )

        # Get the command object to test
        self.cmd = network_qos_rule.ShowNetworkQosRule(self.app, None)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_show_all_options(self):
        arglist = [
            self.new_rule.qos_policy_id,
            self.new_rule.id,
        ]
        verifylist = [
            ('qos_policy', self.new_rule.qos_policy_id),
            ('id', self.new_rule.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.get_qos_bandwidth_limit_rule.assert_called_once_with(
            self.new_rule.id, self.qos_policy.id
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.data), list(data))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import argparse
from collections.abc import Generator
import contextlib
import logging
from typing import Any

import openstack.exceptions
from osc_lib.cli import parseractions
from osc_lib import exceptions

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.network import utils


LOG = logging.getLogger(__name__)

_required_opt_extensions_map = {
    'allowed_address_pairs': 'allowed-address-pairs',
    'dns_domain': 'dns-integration',
    'dns_name': 'dns-integration',
    'extra_dhcp_opts': 'extra_dhcp_opt',
    'qos_policy_id': 'qos',
    'security_groups': 'security-groups',
}


@contextlib.contextmanager
def check_missing_extension_if_error(
    client_manager: Any, attrs: dict[str, Any]
) -> Generator[None, None, None]:
    # If specified option requires extension, then try to
    # find out if it exists. If it does not exist,
    # then an exception with the appropriate message
    # will be thrown from within client.find_extension
    try:
        yield
    except openstack.exceptions.HttpException:
        for opt, ext in _required_opt_extensions_map.items():
            if opt in attrs:
                client_manager.find_extension(ext, ignore_missing=False)
        raise


class NeutronCommandWithExtraArgs(command.Command):
    """Create and Update commands with additional extra properties.

    Extra properties can be passed to the command and are then send to the
    Neutron as given to the command.
    """

    # dict of allowed types
    _allowed_types_dict = {
        'bool': utils.str2bool,
        'dict': utils.str2dict,
        'list': utils.str2list,
        'int': int,
        'str': str,
    }

    def _get_property_converter(self, _property: dict[str, Any]) -> Any:
        if 'type' in _property:
            converter = self._allowed_types_dict.get(_property['type'])
        else:
            converter = str

        if not converter:
            raise exceptions.CommandError(
                _(
                    "Type {property_type} of property {name} is not supported"
                ).format(
                    property_type=_property['type'], name=_property['name']
                )
            )
        return converter

    def _parse_extra_properties(
        self, extra_properties: list[dict[str, Any]] | None
    ) -> dict[str, Any]:
        result: dict[str, Any] = {}
        if extra_properties:
            for _property in extra_properties:
                converter = self._get_property_converter(_property)
                result[_property['name']] = converter(_property['value'])
        return result

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--extra-property',
            metavar='type=<property_type>,name=<property_name>,'
            'value=<property_value>',
            dest='extra_properties',
            action=parseractions.MultiKeyValueAction,
            required_keys=['name', 'value'],
            optional_keys=['type'],
            help=_(
                "Additional parameters can be passed using this property. "
                "Default type of the extra property is string ('str'), but "
                "other types can be used as well. Available types are: "
                "'dict', 'list', 'str', 'bool', 'int'. "
                "In case of 'list' type, 'value' can be "
                "semicolon-separated list of values. "
                "For 'dict' value is semicolon-separated list of the "
                "key:value pairs."
            ),
        )
        return parser


class NeutronUnsetCommandWithExtraArgs(NeutronCommandWithExtraArgs):
    def _parse_extra_properties(
        self, extra_properties: list[dict[str, Any]] | None
    ) -> dict[str, Any]:
        result: dict[str, Any] = {}
        if extra_properties:
            for _property in extra_properties:
                result[_property['name']] = None

        return result



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)


def _get_columns(item: Any) -> tuple[tuple[str, ...], tuple[str, ...]]:
    hidden_columns = ['location']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, {}, hidden_columns
    )


class CreateMetadefResourceTypeAssociation(command.ShowOne):
    _description = _("Create metadef resource type association")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "namespace",
            metavar="<namespace>",
            help=_(
                "The name of the namespace you want to create the "
                "resource type association in"
            ),
        )
        parser.add_argument(
            "name",
            metavar="<name>",
            help=_("A name of the new resource type"),
        )
        parser.add_argument(
            "--properties-target",
            metavar="<properties_target>",
            help=_(
                "Some resource types allow more than one "
                "key/value pair per instance."
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        image_client = self.app.client_manager.image
        kwargs = {}

        kwargs['namespace'] = parsed_args.namespace
        kwargs['name'] = parsed_args.name

        if parsed_args.properties_target:
            kwargs['properties_target'] = parsed_args.properties_target

        obj = image_client.create_metadef_resource_type_association(
            parsed_args.namespace, **kwargs
        )

        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters={})

        return (display_columns, data)


class DeleteMetadefResourceTypeAssociation(command.Command):
    _description = _("Delete metadef resource type association")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "metadef_namespace",
            metavar="<metadef_namespace>",
            help=_("The name of the namespace whose details you want to see"),
        )
        parser.add_argument(
            "name",
            metavar="<name>",
            nargs="+",
            help=_(
                "The name of the resource type(s) (repeat option to delete"
                "multiple metadef resource type associations)"
            ),
        )
        parser.add_argument(
            "--force",
            dest='force',
            action='store_true',
            default=False,
            help=_(
                "Force delete the resource type association if the"
                "namespace is protected"
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        image_client = self.app.client_manager.image

        result = 0
        for resource_type in parsed_args.name:
            try:
                metadef_namespace = image_client.get_metadef_namespace(
                    parsed_args.metadef_namespace
                )

                kwargs = {}
                is_initially_protected = (
                    True if metadef_namespace.is_protected else False
                )
                if is_initially_protected and parsed_args.force:
                    kwargs['is_protected'] = False

                    image_client.update_metadef_namespace(
                        metadef_namespace.namespace, **kwargs
                    )

                try:
                    image_client.delete_metadef_resource_type_association(
                        resource_type, metadef_namespace, ignore_missing=False
                    )
                finally:
                    if is_initially_protected:
                        kwargs['is_protected'] = True
                        image_client.update_metadef_namespace(
                            metadef_namespace.namespace, **kwargs
                        )

            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete resource type with name or "
                        "ID '%(resource_type)s': %(e)s"
                    ),
                    {'resource_type': resource_type, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.metadef_namespace)
            msg = _(
                "%(result)s of %(total)s resource type failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListMetadefResourceTypeAssociations(command.Lister):
    _description = _("List metadef resource type associations")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "metadef_namespace",
            metavar="<metadef_namespace>",
            help=_("The name of the namespace whose details you want to see"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        image_client = self.app.client_manager.image
        data = image_client.metadef_resource_type_associations(
            parsed_args.metadef_namespace,
        )
        columns = ['Name']
        column_headers = columns
        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                )
                for s in data
            ),
        )



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v3 Service action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common


LOG = logging.getLogger(__name__)


def _format_service(service: Any) -> tuple[tuple[str, ...], Any]:
    columns = (
        'id',
        'name',
        'type',
        'is_enabled',
        'description',
    )
    column_headers = (
        'id',
        'name',
        'type',
        'enabled',
        'description',
    )

    return (
        column_headers,
        utils.get_item_properties(
            service,
            columns,
        ),
    )


class CreateService(command.ShowOne):
    _description = _("Create new service")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'type',
            metavar='<type>',
            help=_('New service type (compute, image, identity, volume, etc)'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('New service name'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New service description'),
        )
        enable_group = parser.add_mutually_exclusive_group()
        enable_group.add_argument(
            '--enable',
            action='store_true',
            dest='is_enabled',
            default=True,
            help=_('Enable service (default)'),
        )
        enable_group.add_argument(
            '--disable',
            action='store_false',
            dest='is_enabled',
            default=True,
            help=_('Disable service'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        service = identity_client.create_service(
            name=parsed_args.name,
            type=parsed_args.type,
            description=parsed_args.description,
            is_enabled=parsed_args.is_enabled,
        )

        return _format_service(service)


class DeleteService(command.Command):
    _description = _("Delete service(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'service',
            metavar='<service>',
            nargs='+',
            help=_('Service(s) to delete (type, name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )
        result = 0
        for i in parsed_args.service:
            try:
                service = common.find_service_sdk(identity_client, i)
                identity_client.delete_service(service.id)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete consumer with type, "
                        "name or ID '%(service)s': %(e)s"
                    ),
                    {'service': i, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.service)
            msg = _("%(result)s of %(total)s services failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListService(command.Lister):
    _description = _("List services")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        columns: tuple[str, ...] = ('id', 'name', 'type')
        column_headers: tuple[str, ...] = ('ID', 'Name', 'Type')
        if parsed_args.long:
            columns += ('description', 'is_enabled')
            column_headers += ('Description', 'Enabled')

        data = identity_client.services()

        return (
            column_headers,
            (utils.get_item_properties(s, columns) for s in data),
        )


class SetService(command.Command):
    _description = _("Set service properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'service',
            metavar='<service>',
            help=_('Service to modify (type, name or ID)'),
        )
        parser.add_argument(
            '--type',
            metavar='<type>',
            help=_('New service type (compute, image, identity, volume, etc)'),
        )
        parser.add_argument(
            '--name',
            metavar='<service-name>',
            help=_('New service name'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New service description'),
        )
        enable_group = parser.add_mutually_exclusive_group()
        enable_group.add_argument(
            '--enable',
            action='store_true',
            dest='is_enabled',
            default=None,
            help=_('Enable service'),
        )
        enable_group.add_argument(
            '--disable',
            action='store_false',
            dest='is_enabled',
            default=None,
            help=_('Disable service'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        service = common.find_service_sdk(identity_client, parsed_args.service)
        kwargs = {}
        if parsed_args.type:
            kwargs['type'] = parsed_args.type
        if parsed_args.name:
            kwargs['name'] = parsed_args.name
        if parsed_args.description:
            kwargs['description'] = parsed_args.description
        if parsed_args.is_enabled is not None:
            kwargs['is_enabled'] = parsed_args.is_enabled

        identity_client.update_service(service.id, **kwargs)


class ShowService(command.ShowOne):
    _description = _("Display service details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'service',
            metavar='<service>',
            help=_('Service to display (type, name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        service = common.find_service_sdk(identity_client, parsed_args.service)

        return _format_service(service)



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v3 Endpoint action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common


LOG = logging.getLogger(__name__)


def _format_endpoint(
    endpoint: Any, service: Any
) -> tuple[tuple[str, ...], Any]:
    columns = (
        'is_enabled',
        'id',
        'interface',
        'region_id',
        'region_id',
        'service_id',
        'url',
    )
    column_headers = (
        'enabled',
        'id',
        'interface',
        'region',
        'region_id',
        'service_id',
        'url',
        'service_name',
        'service_type',
    )

    data = utils.get_item_properties(endpoint, columns)
    data += (getattr(service, 'name', ''), service.type)
    return column_headers, data


class AddProjectToEndpoint(command.Command):
    _description = _("Associate a project to an endpoint")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'endpoint',
            metavar='<endpoint>',
            help=_(
                'Endpoint to associate with specified project (name or ID)'
            ),
        )
        parser.add_argument(
            'project',
            metavar='<project>',
            help=_('Project to associate with specified endpoint name or ID)'),
        )
        common.add_project_domain_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.identity

        endpoint = utils.find_resource(client.endpoints, parsed_args.endpoint)

        project = common.find_project(
            client, parsed_args.project, parsed_args.project_domain
        )

        client.endpoint_filter.add_endpoint_to_project(
            project=project.id, endpoint=endpoint.id
        )


class CreateEndpoint(command.ShowOne):
    _description = _("Create new endpoint")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'service',
            metavar='<service>',
            help=_('Service to be associated with new endpoint (name or ID)'),
        )
        parser.add_argument(
            'interface',
            metavar='<interface>',
            choices=['admin', 'public', 'internal'],
            help=_('New endpoint interface type (admin, public or internal)'),
        )
        parser.add_argument(
            'url',
            metavar='<url>',
            help=_('New endpoint URL'),
        )
        parser.add_argument(
            '--region',
            metavar='<region-id>',
            help=_('New endpoint region ID'),
        )
        enable_group = parser.add_mutually_exclusive_group()
        enable_group.add_argument(
            '--enable',
            dest='enabled',
            action='store_true',
            default=True,
            help=_('Enable endpoint (default)'),
        )
        enable_group.add_argument(
            '--disable',
            dest='enabled',
            action='store_false',
            help=_('Disable endpoint'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )
        service = common.find_service_sdk(identity_client, parsed_args.service)

        kwargs = {}

        kwargs['service_id'] = service.id
        kwargs['url'] = parsed_args.url
        kwargs['interface'] = parsed_args.interface
        kwargs['is_enabled'] = parsed_args.enabled

        if parsed_args.region:
            region = identity_client.get_region(parsed_args.region)
            kwargs['region_id'] = region.id

        endpoint = identity_client.create_endpoint(**kwargs)

        return _format_endpoint(endpoint, service=service)


class DeleteEndpoint(command.Command):
    _description = _("Delete endpoint(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'endpoint',
            metavar='<endpoint-id>',
            nargs='+',
            help=_('Endpoint(s) to delete (ID only)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )
        result = 0
        for i in parsed_args.endpoint:
            try:
                endpoint_id = identity_client.find_endpoint(
                    i, ignore_missing=False
                ).id
                identity_client.delete_endpoint(endpoint_id)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete endpoint with "
                        "ID '%(endpoint)s': %(e)s"
                    ),
                    {'endpoint': i, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.endpoint)
            msg = _("%(result)s of %(total)s endpoints failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListEndpoint(command.Lister):
    _description = _("List endpoints")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--service',
            metavar='<service>',
            help=_('Filter by service (type, name or ID)'),
        )
        parser.add_argument(
            '--interface',
            metavar='<interface>',
            choices=['admin', 'public', 'internal'],
            help=_('Filter by interface type (admin, public or internal)'),
        )
        parser.add_argument(
            '--region',
            metavar='<region-id>',
            help=_('Filter by region ID'),
        )
        list_group = parser.add_mutually_exclusive_group()
        list_group.add_argument(
            '--endpoint',
            metavar='<endpoint-group>',
            help=_('Endpoint to list filters'),
        )
        list_group.add_argument(
            '--project',
            metavar='<project>',
            help=_('Project to list filters (name or ID)'),
        )
        common.add_project_domain_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        endpoint = None
        if parsed_args.endpoint:
            endpoint = identity_client.find_endpoint(
                parsed_args.endpoint, ignore_missing=False
            )

        project_domain_id = None
        if parsed_args.project_domain:
            project_domain_id = common._find_sdk_id(
                identity_client.find_domain,
                name_or_id=parsed_args.project_domain,
            )

        project_id = None
        if parsed_args.project:
            project_id = common._find_sdk_id(
                identity_client.find_project,
                name_or_id=common._get_token_resource(
                    identity_client, 'project', parsed_args.project
                ),
                domain_id=project_domain_id,
            )

        if endpoint:
            column_headers: tuple[str, ...] = ('ID', 'Name')
            columns: tuple[str, ...] = ('id', 'name')
            data = identity_client.endpoint_projects(endpoint=endpoint.id)
        else:
            column_headers = (
                'ID',
                'Region',
                'Service Name',
                'Service Type',
                'Enabled',
                'Interface',
                'URL',
            )
            columns = (
                'id',
                'region_id',
                'service_name',
                'service_type',
                'is_enabled',
                'interface',
                'url',
            )
            kwargs = {}
            if parsed_args.service:
                service = common.find_service_sdk(
                    identity_client, parsed_args.service
                )
                kwargs['service_id'] = service.id
            if parsed_args.interface:
                kwargs['interface'] = parsed_args.interface
            if parsed_args.region:
                region = identity_client.get_region(parsed_args.region)
                kwargs['region_id'] = region.id

            if project_id:
                data = list(
                    identity_client.project_endpoints(project=project_id)
                )
            else:
                data = list(identity_client.endpoints(**kwargs))

            for ep in data:
                service = identity_client.find_service(
                    ep.service_id, ignore_missing=False
                )
                ep.service_name = getattr(service, 'name', '')
                ep.service_type = service.type

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class RemoveProjectFromEndpoint(command.Command):
    _description = _("Dissociate a project from an endpoint")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'endpoint',
            metavar='<endpoint>',
            help=_(
                'Endpoint to dissociate from specified project (name or ID)'
            ),
        )
        parser.add_argument(
            'project',
            metavar='<project>',
            help=_(
                'Project to dissociate from specified endpoint name or ID)'
            ),
        )
        common.add_project_domain_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.identity

        endpoint = utils.find_resource(client.endpoints, parsed_args.endpoint)

        project = common.find_project(
            client, parsed_args.project, parsed_args.project_domain
        )

        client.endpoint_filter.delete_endpoint_from_project(
            project=project.id, endpoint=endpoint.id
        )


class SetEndpoint(command.Command):
    _description = _("Set endpoint properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'endpoint',
            metavar='<endpoint-id>',
            help=_('Endpoint to modify (ID only)'),
        )
        parser.add_argument(
            '--region',
            metavar='<region-id>',
            help=_('New endpoint region ID'),
        )
        parser.add_argument(
            '--interface',
            metavar='<interface>',
            choices=['admin', 'public', 'internal'],
            help=_('New endpoint interface type (admin, public or internal)'),
        )
        parser.add_argument(
            '--url',
            metavar='<url>',
            help=_('New endpoint URL'),
        )
        parser.add_argument(
            '--service',
            metavar='<service>',
            help=_('New endpoint service (name or ID)'),
        )
        enable_group = parser.add_mutually_exclusive_group()
        enable_group.add_argument(
            '--enable',
            dest='enabled',
            action='store_true',
            help=_('Enable endpoint'),
        )
        enable_group.add_argument(
            '--disable',
            dest='disabled',
            action='store_true',
            help=_('Disable endpoint'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )
        endpoint = identity_client.find_endpoint(
            parsed_args.endpoint, ignore_missing=False
        )

        kwargs = {}

        if parsed_args.service:
            service = common.find_service_sdk(
                identity_client, parsed_args.service
            )
            kwargs['service_id'] = service.id

        if parsed_args.enabled:
            kwargs['is_enabled'] = True
        if parsed_args.disabled:
            kwargs['is_enabled'] = False

        if parsed_args.url:
            kwargs['url'] = parsed_args.url

        if parsed_args.interface:
            kwargs['interface'] = parsed_args.interface

        if parsed_args.region:
            kwargs['region_id'] = parsed_args.region

        identity_client.update_endpoint(
            endpoint.id,
            **kwargs,
        )


class ShowEndpoint(command.ShowOne):
    _description = _("Display endpoint details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'endpoint',
            metavar='<endpoint>',
            help=_(
                'Endpoint to display (endpoint ID, service ID,'
                ' service name, service type)'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )
        endpoint = identity_client.find_endpoint(
            parsed_args.endpoint, ignore_missing=False
        )

        service = common.find_service_sdk(identity_client, endpoint.service_id)

        return _format_endpoint(endpoint, service)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from osc_lib import exceptions

from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes
from openstackclient.volume.v3 import block_storage_cluster


class TestBlockStorageCluster(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        # Get a shortcut to the BlockStorageClusterManager Mock
        self.cluster_mock = self.volume_client.clusters
        self.cluster_mock.reset_mock()


class TestBlockStorageClusterList(TestBlockStorageCluster):
    # The cluster to be listed
    fake_clusters = volume_fakes.create_clusters()

    def setUp(self):
        super().setUp()

        self.cluster_mock.list.return_value = self.fake_clusters

        # Get the command object to test
        self.cmd = block_storage_cluster.ListBlockStorageCluster(
            self.app, None
        )

    def test_cluster_list(self):
        self.set_volume_api_version('3.7')

        arglist = []
        verifylist = [
            ('cluster', None),
            ('binary', None),
            ('is_up', None),
            ('is_disabled', None),
            ('num_hosts', None),
            ('num_down_hosts', None),
            ('long', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        expected_columns = ('Name', 'Binary', 'State', 'Status')
        expected_data = tuple(
            (
                cluster.name,
                cluster.binary,
                cluster.state,
                cluster.status,
            )
            for cluster in self.fake_clusters
        )
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(expected_columns, columns)
        self.assertEqual(expected_data, tuple(data))

        # checking if proper call was made to list clusters
        self.cluster_mock.list.assert_called_with(
            name=None,
            binary=None,
            is_up=None,
            disabled=None,
            num_hosts=None,
            num_down_hosts=None,
            detailed=False,
        )

    def test_cluster_list_with_full_options(self):
        self.set_volume_api_version('3.7')

        arglist = [
            '--cluster',
            'foo',
            '--binary',
            'bar',
            '--up',
            '--disabled',
            '--num-hosts',
            '5',
            '--num-down-hosts',
            '0',
            '--long',
        ]
        verifylist = [
            ('cluster', 'foo'),
            ('binary', 'bar'),
            ('is_up', True),
            ('is_disabled', True),
            ('num_hosts', 5),
            ('num_down_hosts', 0),
            ('long', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        expected_columns = (
            'Name',
            'Binary',
            'State',
            'Status',
            'Num Hosts',
            'Num Down Hosts',
            'Last Heartbeat',
            'Disabled Reason',
            'Created At',
            'Updated At',
        )
        expected_data = tuple(
            (
                cluster.name,
                cluster.binary,
                cluster.state,
                cluster.status,
                cluster.num_hosts,
                cluster.num_down_hosts,
                cluster.last_heartbeat,
                cluster.disabled_reason,
                cluster.created_at,
                cluster.updated_at,
            )
            for cluster in self.fake_clusters
        )
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(expected_columns, columns)
        self.assertEqual(expected_data, tuple(data))

        # checking if proper call was made to list clusters
        self.cluster_mock.list.assert_called_with(
            name='foo',
            binary='bar',
            is_up=True,
            disabled=True,
            num_hosts=5,
            num_down_hosts=0,
            detailed=True,
        )

    def test_cluster_list_pre_v37(self):
        self.set_volume_api_version('3.6')

        arglist = []
        verifylist = [
            ('cluster', None),
            ('binary', None),
            ('is_up', None),
            ('is_disabled', None),
            ('num_hosts', None),
            ('num_down_hosts', None),
            ('long', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.7 or greater is required', str(exc)
        )


class TestBlockStorageClusterSet(TestBlockStorageCluster):
    cluster = volume_fakes.create_one_cluster()
    columns = (
        'Name',
        'Binary',
        'State',
        'Status',
        'Disabled Reason',
        'Hosts',
        'Down Hosts',
        'Last Heartbeat',
        'Created At',
        'Updated At',
        'Replication Status',
        'Frozen',
        'Active Backend ID',
    )
    data = (
        cluster.name,
        cluster.binary,
        cluster.state,
        cluster.status,
        cluster.disabled_reason,
        cluster.num_hosts,
        cluster.num_down_hosts,
        cluster.last_heartbeat,
        cluster.created_at,
        cluster.updated_at,
        cluster.replication_status,
        cluster.frozen,
        cluster.active_backend_id,
    )

    def setUp(self):
        super().setUp()

        self.cluster_mock.update.return_value = self.cluster

        self.cmd = block_storage_cluster.SetBlockStorageCluster(self.app, None)

    def test_cluster_set(self):
        self.set_volume_api_version('3.7')

        arglist = [
            '--enable',
            self.cluster.name,
        ]
        verifylist = [
            ('cluster', self.cluster.name),
            ('binary', 'cinder-volume'),
            ('disabled', False),
            ('disabled_reason', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, tuple(data))

        self.cluster_mock.update.assert_called_once_with(
            self.cluster.name,
            'cinder-volume',
            disabled=False,
            disabled_reason=None,
        )

    def test_cluster_set_disable_with_reason(self):
        self.set_volume_api_version('3.7')

        arglist = [
            '--binary',
            self.cluster.binary,
            '--disable',
            '--disable-reason',
            'foo',
            self.cluster.name,
        ]
        verifylist = [
            ('cluster', self.cluster.name),
            ('binary', self.cluster.binary),
            ('disabled', True),
            ('disabled_reason', 'foo'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, tuple(data))

        self.cluster_mock.update.assert_called_once_with(
            self.cluster.name,
            self.cluster.binary,
            disabled=True,
            disabled_reason='foo',
        )

    def test_cluster_set_only_with_disable_reason(self):
        self.set_volume_api_version('3.7')

        arglist = [
            '--disable-reason',
            'foo',
            self.cluster.name,
        ]
        verifylist = [
            ('cluster', self.cluster.name),
            ('binary', 'cinder-volume'),
            ('disabled', None),
            ('disabled_reason', 'foo'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            "Cannot specify --disable-reason without --disable", str(exc)
        )

    def test_cluster_set_enable_with_disable_reason(self):
        self.set_volume_api_version('3.7')

        arglist = [
            '--enable',
            '--disable-reason',
            'foo',
            self.cluster.name,
        ]
        verifylist = [
            ('cluster', self.cluster.name),
            ('binary', 'cinder-volume'),
            ('disabled', False),
            ('disabled_reason', 'foo'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            "Cannot specify --disable-reason without --disable", str(exc)
        )

    def test_cluster_set_pre_v37(self):
        self.set_volume_api_version('3.6')

        arglist = [
            '--enable',
            self.cluster.name,
        ]
        verifylist = [
            ('cluster', self.cluster.name),
            ('binary', 'cinder-volume'),
            ('disabled', False),
            ('disabled_reason', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.7 or greater is required', str(exc)
        )


class TestBlockStorageClusterShow(TestBlockStorageCluster):
    cluster = volume_fakes.create_one_cluster()
    columns = (
        'Name',
        'Binary',
        'State',
        'Status',
        'Disabled Reason',
        'Hosts',
        'Down Hosts',
        'Last Heartbeat',
        'Created At',
        'Updated At',
        'Replication Status',
        'Frozen',
        'Active Backend ID',
    )
    data = (
        cluster.name,
        cluster.binary,
        cluster.state,
        cluster.status,
        cluster.disabled_reason,
        cluster.num_hosts,
        cluster.num_down_hosts,
        cluster.last_heartbeat,
        cluster.created_at,
        cluster.updated_at,
        cluster.replication_status,
        cluster.frozen,
        cluster.active_backend_id,
    )

    def setUp(self):
        super().setUp()

        self.cluster_mock.show.return_value = self.cluster

        self.cmd = block_storage_cluster.ShowBlockStorageCluster(
            self.app, None
        )

    def test_cluster_show(self):
        self.set_volume_api_version('3.7')

        arglist = [
            '--binary',
            self.cluster.binary,
            self.cluster.name,
        ]
        verifylist = [
            ('cluster', self.cluster.name),
            ('binary', self.cluster.binary),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, tuple(data))

        self.cluster_mock.show.assert_called_once_with(
            self.cluster.name,
            binary=self.cluster.binary,
        )

    def test_cluster_show_pre_v37(self):
        self.set_volume_api_version('3.6')

        arglist = [
            '--binary',
            self.cluster.binary,
            self.cluster.name,
        ]
        verifylist = [
            ('cluster', self.cluster.name),
            ('binary', self.cluster.binary),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.7 or greater is required', str(exc)
        )



#   Copyright 2016 IBM Corporation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import uuid

from openstackclient.compute.v2 import host
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes
from openstackclient.tests.unit import fakes
from openstackclient.tests.unit import utils as tests_utils


def _generate_fake_host():
    return {
        'service_id': 1,
        'host': 'host1',
        'uuid': 'host-id-' + uuid.uuid4().hex,
        'vcpus': 10,
        'memory_mb': 100,
        'local_gb': 100,
        'vcpus_used': 5,
        'memory_mb_used': 50,
        'local_gb_used': 10,
        'hypervisor_type': 'xen',
        'hypervisor_version': 1,
        'hypervisor_hostname': 'devstack1',
        'free_ram_mb': 50,
        'free_disk_gb': 50,
        'current_workload': 10,
        'running_vms': 1,
        'cpu_info': '',
        'disk_available_least': 1,
        'host_ip': '10.10.10.10',
        'supported_instances': '',
        'metrics': '',
        'pci_stats': '',
        'extra_resources': '',
        'stats': '',
        'numa_topology': '',
        'ram_allocation_ratio': 1.0,
        'cpu_allocation_ratio': 1.0,
        'zone': 'zone-' + uuid.uuid4().hex,
        'host_name': 'name-' + uuid.uuid4().hex,
        'service': 'service-' + uuid.uuid4().hex,
        'cpu': 4,
        'disk_gb': 100,
        'project': 'project-' + uuid.uuid4().hex,
    }


class TestHostList(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self._host = _generate_fake_host()
        self.columns = ('Host Name', 'Service', 'Zone')
        self.data = [
            (
                self._host['host_name'],
                self._host['service'],
                self._host['zone'],
            )
        ]

        self.compute_client.get.return_value = fakes.FakeResponse(
            data={'hosts': [self._host]}
        )
        self.cmd = host.ListHost(self.app, None)

    def test_host_list_no_option(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.get.assert_called_with(
            '/os-hosts', microversion='2.1'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_host_list_with_option(self):
        arglist = [
            '--zone',
            self._host['zone'],
        ]
        verifylist = [
            ('zone', self._host['zone']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.get.assert_called_with(
            '/os-hosts', microversion='2.1'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))


class TestHostSet(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self._host = _generate_fake_host()
        self.compute_client.put.return_value = fakes.FakeResponse()

        self.cmd = host.SetHost(self.app, None)

    def test_host_set_no_option(self):
        arglist = [
            self._host['host'],
        ]
        verifylist = [
            ('host', self._host['host']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)
        self.compute_client.put.assert_not_called()

    def test_host_set(self):
        arglist = [
            '--enable',
            '--disable-maintenance',
            self._host['host'],
        ]
        verifylist = [
            ('enable', True),
            ('enable_maintenance', False),
            ('host', self._host['host']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)
        self.compute_client.put.assert_called_with(
            f'/os-hosts/{self._host["host"]}',
            json={
                'maintenance_mode': 'disable',
                'status': 'enable',
            },
            microversion='2.1',
        )


class TestHostShow(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self._host = _generate_fake_host()

        self.columns = (
            'Host',
            'Project',
            'CPU',
            'Memory MB',
            'Disk GB',
        )
        self.data = [
            (
                self._host['host'],
                self._host['project'],
                self._host['cpu'],
                self._host['memory_mb'],
                self._host['disk_gb'],
            )
        ]

        self.compute_client.get.return_value = fakes.FakeResponse(
            data={
                'host': [
                    {
                        'resource': {
                            'host': self._host['host'],
                            'project': self._host['project'],
                            'cpu': self._host['cpu'],
                            'memory_mb': self._host['memory_mb'],
                            'disk_gb': self._host['disk_gb'],
                        }
                    }
                ],
            }
        )

        self.cmd = host.ShowHost(self.app, None)

    def test_host_show_no_option(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_host_show_with_option(self):
        arglist = [
            self._host['host_name'],
        ]
        verifylist = [
            ('host', self._host['host_name']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.get.assert_called_with(
            '/os-hosts/' + self._host['host_name'], microversion='2.1'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Extension action implementations"""

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)


def _get_extension_columns(
    item: Any,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    column_map = {
        'updated': 'updated_at',
    }
    hidden_columns = ['id', 'links', 'location']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns
    )


class ListExtension(command.Lister):
    _description = _("List API extensions")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--compute',
            action='store_true',
            default=False,
            help=_('List extensions for the Compute API'),
        )
        parser.add_argument(
            '--identity',
            action='store_true',
            default=False,
            help=_('List extensions for the Identity API'),
        )
        parser.add_argument(
            '--network',
            action='store_true',
            default=False,
            help=_('List extensions for the Network API'),
        )
        parser.add_argument(
            '--volume',
            action='store_true',
            default=False,
            help=_('List extensions for the Block Storage API'),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        columns: tuple[str, ...] = ('Name', 'Alias', 'Description')
        if parsed_args.long:
            columns += ('Namespace', 'Updated At', 'Links')

        data = []

        # by default we want to show everything, unless the
        # user specifies one or more of the APIs to show
        # for now, only identity and compute are supported.
        show_all = (
            not parsed_args.identity
            and not parsed_args.compute
            and not parsed_args.volume
            and not parsed_args.network
        )

        if parsed_args.identity or show_all:
            identity_client = self.app.client_manager.identity
            try:
                data += identity_client.extensions.list()
            except Exception:
                message = _("Extensions list not supported by Identity API")
                LOG.warning(message)

        if parsed_args.compute or show_all:
            compute_client = self.app.client_manager.compute
            try:
                data += compute_client.extensions()
            except Exception:
                message = _("Extensions list not supported by Compute API")
                LOG.warning(message)

        if parsed_args.volume or show_all:
            volume_client = self.app.client_manager.sdk_connection.volume
            try:
                data += volume_client.extensions()
            except Exception:
                message = _(
                    "Extensions list not supported by Block Storage API"
                )
                LOG.warning(message)

        if parsed_args.network or show_all:
            network_client = self.app.client_manager.network
            try:
                data += network_client.extensions()
            except Exception:
                message = _(
                    "Failed to retrieve extensions list from Network API"
                )
                LOG.warning(message)

        extension_tuples = (
            utils.get_item_properties(
                s,
                columns,
            )
            for s in data
        )

        return (columns, extension_tuples)


class ShowExtension(command.ShowOne):
    _description = _("Show API extension")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'extension',
            metavar='<extension>',
            help=_(
                'Extension to display. '
                'Currently, only network extensions are supported. '
                '(Name or Alias)'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network

        extension = client.find_extension(
            parsed_args.extension,
            ignore_missing=False,
        )

        display_columns, columns = _get_extension_columns(extension)
        data = utils.get_dict_properties(extension, columns)
        return display_columns, data



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#


"""Identity v3 Protocols actions implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


def _format_protocol(protocol: Any) -> tuple[tuple[str, ...], Any]:
    columns = ('name', 'idp_id', 'mapping_id')
    column_headers = ('id', 'identity_provider', 'mapping')
    return (
        column_headers,
        utils.get_item_properties(protocol, columns),
    )


class CreateProtocol(command.ShowOne):
    _description = _("Create new federation protocol")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'federation_protocol',
            metavar='<name>',
            help=_(
                'New federation protocol name (must be unique '
                'per identity provider)'
            ),
        )
        parser.add_argument(
            '--identity-provider',
            metavar='<identity-provider>',
            required=True,
            help=_(
                'Identity provider that will support the new federation '
                ' protocol (name or ID) (required)'
            ),
        )
        parser.add_argument(
            '--mapping',
            metavar='<mapping>',
            required=True,
            help=_('Mapping that is to be used (name or ID) (required)'),
        )

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        protocol = identity_client.create_federation_protocol(
            name=parsed_args.federation_protocol,
            idp_id=parsed_args.identity_provider,
            mapping_id=parsed_args.mapping,
        )

        return _format_protocol(protocol)


class DeleteProtocol(command.Command):
    _description = _("Delete federation protocol(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'federation_protocol',
            metavar='<federation-protocol>',
            nargs='+',
            help=_('Federation protocol(s) to delete (name or ID)'),
        )
        parser.add_argument(
            '--identity-provider',
            metavar='<identity-provider>',
            required=True,
            help=_(
                'Identity provider that supports <federation-protocol> '
                '(name or ID) (required)'
            ),
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        result = 0
        for i in parsed_args.federation_protocol:
            try:
                identity_client.delete_federation_protocol(
                    idp_id=parsed_args.identity_provider,
                    protocol=i,
                    ignore_missing=False,
                )
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete federation protocol "
                        "with name or ID '%(protocol)s': %(e)s"
                    ),
                    {'protocol': i, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.federation_protocol)
            msg = _(
                "%(result)s of %(total)s federation protocols failed"
                " to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListProtocols(command.Lister):
    _description = _("List federation protocols")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--identity-provider',
            metavar='<identity-provider>',
            required=True,
            help=_('Identity provider to list (name or ID) (required)'),
        )

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        protocols = identity_client.federation_protocols(
            parsed_args.identity_provider
        )
        columns = ('id', 'mapping')
        response_attributes = ('id', 'mapping_id')
        items = [
            utils.get_item_properties(s, response_attributes)
            for s in protocols
        ]
        return (columns, items)


class SetProtocol(command.Command):
    _description = _("Set federation protocol properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'federation_protocol',
            metavar='<name>',
            help=_('Federation protocol to modify (name or ID)'),
        )
        parser.add_argument(
            '--identity-provider',
            metavar='<identity-provider>',
            required=True,
            help=_(
                'Identity provider that supports <federation-protocol> '
                '(name or ID) (required)'
            ),
        )
        parser.add_argument(
            '--mapping',
            metavar='<mapping>',
            help=_('Mapping that is to be used (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        kwargs = {'idp_id': parsed_args.identity_provider}
        if parsed_args.federation_protocol:
            kwargs['name'] = parsed_args.federation_protocol
        if parsed_args.mapping:
            kwargs['mapping_id'] = parsed_args.mapping

        protocol = identity_client.update_federation_protocol(**kwargs)
        return _format_protocol(protocol)


class ShowProtocol(command.ShowOne):
    _description = _("Display federation protocol details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'federation_protocol',
            metavar='<federation-protocol>',
            help=_('Federation protocol to display (name or ID)'),
        )
        parser.add_argument(
            '--identity-provider',
            metavar='<identity-provider>',
            required=True,
            help=_(
                'Identity provider that supports <federation-protocol> '
                '(name or ID) (required)'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        protocol = identity_client.get_federation_protocol(
            idp_id=parsed_args.identity_provider,
            protocol=parsed_args.federation_protocol,
        )
        return _format_protocol(protocol)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""IP Availability Info implementations"""

import argparse
from collections.abc import Iterable, Sequence
from typing import Any

from osc_lib.cli import format_columns
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common

_formatters = {
    'subnet_ip_availability': format_columns.ListDictColumn,
}


def _get_columns(item: Any) -> tuple[tuple[str, ...], tuple[str, ...]]:
    hidden_columns = ['id', 'name', 'location', 'tenant_id']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, {}, hidden_columns
    )


# TODO(ankur-gupta-f): Use the SDK resource mapped attribute names once
# the OSC minimum requirements include SDK 1.0.
class ListIPAvailability(command.Lister):
    _description = _("List IP availability for network")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--ip-version',
            type=int,
            default=4,
            choices=[4, 6],
            metavar='<ip-version>',
            dest='ip_version',
            help=_(
                "List only IP availability with the specified IP version "
                "networks (4 or 6, default is 4)"
            ),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                "List only IP availability with the specified project "
                "(name or ID)"
            ),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network

        columns = (
            'network_id',
            'network_name',
            'total_ips',
            'used_ips',
        )
        column_headers = (
            'Network ID',
            'Network Name',
            'Total IPs',
            'Used IPs',
        )

        filters = {}
        if parsed_args.ip_version:
            filters['ip_version'] = parsed_args.ip_version

        if parsed_args.project:
            identity_client = self.app.client_manager.identity
            project_id = identity_common.find_project(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            ).id
            filters['project_id'] = project_id
        data = client.network_ip_availabilities(**filters)
        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                )
                for s in data
            ),
        )


class ShowIPAvailability(command.ShowOne):
    _description = _("Show network IP availability details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'network',
            metavar="<network>",
            help=_("Show IP availability for a specific network (name or ID)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        network_id = client.find_network(
            parsed_args.network, ignore_missing=False
        ).id
        obj = client.find_network_ip_availability(
            network_id, ignore_missing=False
        )
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters=_formatters)
        return (display_columns, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest.mock import call

from osc_lib import exceptions

from openstackclient.network.v2 import address_scope
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes_v3
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestAddressScope(network_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        # Get a shortcut to the ProjectManager Mock
        self.projects_mock = self.identity_client.projects
        # Get a shortcut to the DomainManager Mock
        self.domains_mock = self.identity_client.domains


class TestCreateAddressScope(TestAddressScope):
    project = identity_fakes_v3.FakeProject.create_one_project()
    domain = identity_fakes_v3.FakeDomain.create_one_domain()
    # The new address scope created.
    new_address_scope = network_fakes.create_one_address_scope(
        attrs={
            'project_id': project.id,
        }
    )
    columns = ('id', 'ip_version', 'name', 'project_id', 'shared')
    data = (
        new_address_scope.id,
        new_address_scope.ip_version,
        new_address_scope.name,
        new_address_scope.project_id,
        new_address_scope.is_shared,
    )

    def setUp(self):
        super().setUp()
        self.network_client.create_address_scope.return_value = (
            self.new_address_scope
        )

        # Get the command object to test
        self.cmd = address_scope.CreateAddressScope(self.app, None)

        self.projects_mock.get.return_value = self.project
        self.domains_mock.get.return_value = self.domain

    def test_create_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_default_options(self):
        arglist = [
            self.new_address_scope.name,
        ]
        verifylist = [
            ('project', None),
            ('ip_version', self.new_address_scope.ip_version),
            ('name', self.new_address_scope.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_address_scope.assert_called_once_with(
            **{
                'ip_version': self.new_address_scope.ip_version,
                'name': self.new_address_scope.name,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_all_options(self):
        arglist = [
            '--ip-version',
            str(self.new_address_scope.ip_version),
            '--share',
            '--project',
            self.project.name,
            '--project-domain',
            self.domain.name,
            self.new_address_scope.name,
        ]
        verifylist = [
            ('ip_version', self.new_address_scope.ip_version),
            ('share', True),
            ('project', self.project.name),
            ('project_domain', self.domain.name),
            ('name', self.new_address_scope.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_address_scope.assert_called_once_with(
            **{
                'ip_version': self.new_address_scope.ip_version,
                'shared': True,
                'project_id': self.project.id,
                'name': self.new_address_scope.name,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_no_share(self):
        arglist = [
            '--no-share',
            self.new_address_scope.name,
        ]
        verifylist = [
            ('no_share', True),
            ('name', self.new_address_scope.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_address_scope.assert_called_once_with(
            **{
                'ip_version': self.new_address_scope.ip_version,
                'shared': False,
                'name': self.new_address_scope.name,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestDeleteAddressScope(TestAddressScope):
    # The address scope to delete.
    _address_scopes = network_fakes.create_address_scopes(count=2)

    def setUp(self):
        super().setUp()
        self.network_client.delete_address_scope.return_value = None
        self.network_client.find_address_scope = (
            network_fakes.get_address_scopes(
                address_scopes=self._address_scopes
            )
        )

        # Get the command object to test
        self.cmd = address_scope.DeleteAddressScope(self.app, None)

    def test_address_scope_delete(self):
        arglist = [
            self._address_scopes[0].name,
        ]
        verifylist = [
            ('address_scope', [self._address_scopes[0].name]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network_client.find_address_scope.assert_called_once_with(
            self._address_scopes[0].name, ignore_missing=False
        )
        self.network_client.delete_address_scope.assert_called_once_with(
            self._address_scopes[0]
        )
        self.assertIsNone(result)

    def test_multi_address_scopes_delete(self):
        arglist = []
        verifylist = []

        for a in self._address_scopes:
            arglist.append(a.name)
        verifylist = [
            ('address_scope', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        calls = []
        for a in self._address_scopes:
            calls.append(call(a))
        self.network_client.delete_address_scope.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_multi_address_scopes_delete_with_exception(self):
        arglist = [
            self._address_scopes[0].name,
            'unexist_address_scope',
        ]
        verifylist = [
            (
                'address_scope',
                [self._address_scopes[0].name, 'unexist_address_scope'],
            ),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_result = [self._address_scopes[0], exceptions.CommandError]
        self.network_client.find_address_scope.side_effect = find_mock_result

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 address scopes failed to delete.', str(e))

        self.network_client.find_address_scope.assert_any_call(
            self._address_scopes[0].name, ignore_missing=False
        )
        self.network_client.find_address_scope.assert_any_call(
            'unexist_address_scope', ignore_missing=False
        )
        self.network_client.delete_address_scope.assert_called_once_with(
            self._address_scopes[0]
        )


class TestListAddressScope(TestAddressScope):
    # The address scopes to list up.
    address_scopes = network_fakes.create_address_scopes(count=3)
    columns = (
        'ID',
        'Name',
        'IP Version',
        'Shared',
        'Project',
    )
    data = []
    for scope in address_scopes:
        data.append(
            (
                scope.id,
                scope.name,
                scope.ip_version,
                scope.is_shared,
                scope.project_id,
            )
        )

    def setUp(self):
        super().setUp()
        self.network_client.address_scopes.return_value = self.address_scopes

        # Get the command object to test
        self.cmd = address_scope.ListAddressScope(self.app, None)

    def test_address_scope_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.address_scopes.assert_called_once_with(**{})
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_address_scope_list_name(self):
        arglist = [
            '--name',
            self.address_scopes[0].name,
        ]
        verifylist = [
            ('name', self.address_scopes[0].name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.address_scopes.assert_called_once_with(
            **{'name': self.address_scopes[0].name}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_address_scope_list_ip_version(self):
        arglist = [
            '--ip-version',
            str(4),
        ]
        verifylist = [
            ('ip_version', 4),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.address_scopes.assert_called_once_with(
            **{'ip_version': 4}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_address_scope_list_project(self):
        project = identity_fakes_v3.FakeProject.create_one_project()
        self.projects_mock.get.return_value = project
        arglist = [
            '--project',
            project.id,
        ]
        verifylist = [
            ('project', project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.address_scopes.assert_called_once_with(
            **{'project_id': project.id}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_address_scope_project_domain(self):
        project = identity_fakes_v3.FakeProject.create_one_project()
        self.projects_mock.get.return_value = project
        arglist = [
            '--project',
            project.id,
            '--project-domain',
            project.domain_id,
        ]
        verifylist = [
            ('project', project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'project_id': project.id}

        self.network_client.address_scopes.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_address_scope_list_share(self):
        arglist = [
            '--share',
        ]
        verifylist = [
            ('share', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.address_scopes.assert_called_once_with(
            **{'is_shared': True}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_address_scope_list_no_share(self):
        arglist = [
            '--no-share',
        ]
        verifylist = [
            ('no_share', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.address_scopes.assert_called_once_with(
            **{'is_shared': False}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))


class TestSetAddressScope(TestAddressScope):
    # The address scope to set.
    _address_scope = network_fakes.create_one_address_scope()

    def setUp(self):
        super().setUp()
        self.network_client.update_address_scope.return_value = None
        self.network_client.find_address_scope.return_value = (
            self._address_scope
        )

        # Get the command object to test
        self.cmd = address_scope.SetAddressScope(self.app, None)

    def test_set_nothing(self):
        arglist = [
            self._address_scope.name,
        ]
        verifylist = [
            ('address_scope', self._address_scope.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {}
        self.network_client.update_address_scope.assert_called_with(
            self._address_scope, **attrs
        )
        self.assertIsNone(result)

    def test_set_name_and_share(self):
        arglist = [
            '--name',
            'new_address_scope',
            '--share',
            self._address_scope.name,
        ]
        verifylist = [
            ('name', 'new_address_scope'),
            ('share', True),
            ('address_scope', self._address_scope.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {
            'name': "new_address_scope",
            'shared': True,
        }
        self.network_client.update_address_scope.assert_called_with(
            self._address_scope, **attrs
        )
        self.assertIsNone(result)

    def test_set_no_share(self):
        arglist = [
            '--no-share',
            self._address_scope.name,
        ]
        verifylist = [
            ('no_share', True),
            ('address_scope', self._address_scope.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {
            'shared': False,
        }
        self.network_client.update_address_scope.assert_called_with(
            self._address_scope, **attrs
        )
        self.assertIsNone(result)


class TestShowAddressScope(TestAddressScope):
    # The address scope to show.
    _address_scope = network_fakes.create_one_address_scope()
    columns = (
        'id',
        'ip_version',
        'name',
        'project_id',
        'shared',
    )
    data = (
        _address_scope.id,
        _address_scope.ip_version,
        _address_scope.name,
        _address_scope.project_id,
        _address_scope.is_shared,
    )

    def setUp(self):
        super().setUp()
        self.network_client.find_address_scope.return_value = (
            self._address_scope
        )

        # Get the command object to test
        self.cmd = address_scope.ShowAddressScope(self.app, None)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_show_all_options(self):
        arglist = [
            self._address_scope.name,
        ]
        verifylist = [
            ('address_scope', self._address_scope.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.find_address_scope.assert_called_once_with(
            self._address_scope.name, ignore_missing=False
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.data), list(data))



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v2 Project action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from keystoneauth1 import exceptions as ks_exc
from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


class CreateProject(command.ShowOne):
    _description = _("Create new project")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<project-name>',
            help=_('New project name'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Project description'),
        )
        enable_group = parser.add_mutually_exclusive_group()
        enable_group.add_argument(
            '--enable',
            action='store_true',
            help=_('Enable project (default)'),
        )
        enable_group.add_argument(
            '--disable',
            action='store_true',
            help=_('Disable project'),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            dest='properties',
            action=parseractions.KeyValueAction,
            help=_(
                'Add a property to <name> '
                '(repeat option to set multiple properties)'
            ),
        )
        parser.add_argument(
            '--or-show',
            action='store_true',
            help=_('Return existing project'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = self.app.client_manager.identity

        enabled = True
        if parsed_args.disable:
            enabled = False
        kwargs = {}
        if parsed_args.properties:
            kwargs.update(parsed_args.properties)

        try:
            project = identity_client.tenants.create(
                parsed_args.name,
                description=parsed_args.description,
                enabled=enabled,
                **kwargs,
            )
        except ks_exc.Conflict:
            if parsed_args.or_show:
                project = utils.find_resource(
                    identity_client.tenants,
                    parsed_args.name,
                )
                LOG.info(_('Returning existing project %s'), project.name)
            else:
                raise

        # TODO(stevemar): Remove the line below when we support multitenancy
        project._info.pop('parent_id', None)
        col_headers, col_data = zip(*sorted(project._info.items()))
        return col_headers, col_data


class DeleteProject(command.Command):
    _description = _(
        "Delete project(s). This command will remove specified "
        "existing project(s) if an active user is authorized to do "
        "this. If there are resources managed by other services "
        "(for example, Nova, Neutron, Cinder) associated with "
        "specified project(s), delete operation will proceed "
        "regardless."
    )

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'projects',
            metavar='<project>',
            nargs="+",
            help=_('Project(s) to delete (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = self.app.client_manager.identity

        errors = 0
        for project in parsed_args.projects:
            try:
                project_obj = utils.find_resource(
                    identity_client.tenants,
                    project,
                )
                identity_client.tenants.delete(project_obj.id)
            except Exception as e:
                errors += 1
                LOG.error(
                    _(
                        "Failed to delete project with "
                        "name or ID '%(project)s': %(e)s"
                    ),
                    {'project': project, 'e': e},
                )

        if errors > 0:
            total = len(parsed_args.projects)
            msg = _("%(errors)s of %(total)s projects failed to delete.") % {
                'errors': errors,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListProject(command.Lister):
    _description = _("List projects")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        parser.add_argument(
            '--sort',
            metavar='<key>[:<direction>]',
            help=_(
                'Sort output by selected keys and directions (asc or desc) '
                '(default: asc), repeat this option to specify multiple '
                'keys and directions.'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        columns: tuple[str, ...] = ('ID', 'Name')
        if parsed_args.long:
            columns += ('Description', 'Enabled')
        data = self.app.client_manager.identity.tenants.list()
        if parsed_args.sort:
            data = utils.sort_items(data, parsed_args.sort)
        return (
            columns,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class SetProject(command.Command):
    _description = _("Set project properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'project',
            metavar='<project>',
            help=_('Project to modify (name or ID)'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Set project name'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Set project description'),
        )
        enable_group = parser.add_mutually_exclusive_group()
        enable_group.add_argument(
            '--enable',
            action='store_true',
            help=_('Enable project'),
        )
        enable_group.add_argument(
            '--disable',
            action='store_true',
            help=_('Disable project'),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            dest='properties',
            action=parseractions.KeyValueAction,
            help=_(
                'Set a project property '
                '(repeat option to set multiple properties)'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = self.app.client_manager.identity

        project = utils.find_resource(
            identity_client.tenants,
            parsed_args.project,
        )

        kwargs = project._info
        if parsed_args.name:
            kwargs['name'] = parsed_args.name
        if parsed_args.description:
            kwargs['description'] = parsed_args.description
        if parsed_args.enable:
            kwargs['enabled'] = True
        if parsed_args.disable:
            kwargs['enabled'] = False
        if parsed_args.properties:
            kwargs.update(parsed_args.properties)
        if 'id' in kwargs:
            del kwargs['id']
        if 'name' in kwargs:
            # Hack around broken Identity API arg names
            kwargs['tenant_name'] = kwargs['name']
            del kwargs['name']

        identity_client.tenants.update(project.id, **kwargs)


class ShowProject(command.ShowOne):
    _description = _("Display project details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'project',
            metavar='<project>',
            help=_('Project to display (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = self.app.client_manager.identity

        info = {}
        try:
            project = utils.find_resource(
                identity_client.tenants,
                parsed_args.project,
            )
            info.update(project._info)
        except ks_exc.Forbidden:
            auth_ref = self.app.client_manager.auth_ref
            if (
                parsed_args.project == auth_ref.project_id
                or parsed_args.project == auth_ref.project_name
            ):
                # Ask for currently auth'ed project so return it
                info = {
                    'id': auth_ref.project_id,
                    'name': auth_ref.project_name,
                    # True because we don't get this far if it is disabled
                    'enabled': True,
                }
            else:
                raise

        # TODO(stevemar): Remove the line below when we support multitenancy
        info.pop('parent_id', None)

        # NOTE(stevemar): Property handling isn't really supported in Keystone
        # and needs a lot of extra handling. Let's reserve the properties that
        # the API has and handle the extra top level properties.
        reserved = ('name', 'id', 'enabled', 'description')
        properties = {}
        for k, v in list(info.items()):
            if k not in reserved:
                # If a key is not in `reserved` it's a property, pop it
                info.pop(k)
                # If a property has been "unset" it's `None`, so don't show it
                if v is not None:
                    properties[k] = v

        info['properties'] = format_columns.DictColumn(properties)
        col_headers, col_data = zip(*sorted(info.items()))
        return col_headers, col_data


class UnsetProject(command.Command):
    _description = _("Unset project properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'project',
            metavar='<project>',
            help=_('Project to modify (name or ID)'),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            dest='properties',
            action='append',
            default=[],
            help=_(
                'Unset a project property '
                '(repeat option to unset multiple properties)'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = self.app.client_manager.identity
        project = utils.find_resource(
            identity_client.tenants,
            parsed_args.project,
        )
        kwargs = project._info
        for key in parsed_args.properties:
            if key in kwargs:
                kwargs[key] = None
        identity_client.tenants.update(project.id, **kwargs)



#   Copyright 2016 Huawei, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from openstack.compute.v2 import server_group as _server_group
from openstack.test import fakes as sdk_fakes
from osc_lib.cli import format_columns
from osc_lib import exceptions

from openstackclient.compute.v2 import server_group
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestServerGroup(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self.fake_server_group = sdk_fakes.generate_fake_resource(
            _server_group.ServerGroup
        )

        self.columns = (
            'id',
            'members',
            'name',
            'policy',
            'project_id',
            'rules',
            'user_id',
        )
        self.data = (
            self.fake_server_group.id,
            format_columns.ListColumn(self.fake_server_group.member_ids),
            self.fake_server_group.name,
            self.fake_server_group.policy,
            self.fake_server_group.project_id,
            format_columns.DictColumn(self.fake_server_group.rules),
            self.fake_server_group.user_id,
        )


class TestServerGroupCreate(TestServerGroup):
    def setUp(self):
        super().setUp()

        self.compute_client.create_server_group.return_value = (
            self.fake_server_group
        )
        self.cmd = server_group.CreateServerGroup(self.app, None)

    def test_server_group_create(self):
        self.set_compute_api_version('2.64')

        arglist = [
            '--policy',
            'anti-affinity',
            'affinity_group',
        ]
        verifylist = [
            ('policy', 'anti-affinity'),
            ('name', 'affinity_group'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.compute_client.create_server_group.assert_called_once_with(
            name=parsed_args.name,
            policy=parsed_args.policy,
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_server_group_create_with_soft_policies(self):
        self.set_compute_api_version('2.64')

        arglist = [
            '--policy',
            'soft-anti-affinity',
            'affinity_group',
        ]
        verifylist = [
            ('policy', 'soft-anti-affinity'),
            ('name', 'affinity_group'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.compute_client.create_server_group.assert_called_once_with(
            name=parsed_args.name,
            policy=parsed_args.policy,
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_server_group_create_with_soft_policies_pre_v215(self):
        self.set_compute_api_version('2.14')

        arglist = [
            '--policy',
            'soft-anti-affinity',
            'affinity_group',
        ]
        verifylist = [
            ('policy', 'soft-anti-affinity'),
            ('name', 'affinity_group'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-compute-api-version 2.15 or greater is required', str(ex)
        )

    def test_server_group_create_with_rules(self):
        self.set_compute_api_version('2.64')

        arglist = [
            '--policy',
            'soft-anti-affinity',
            '--rule',
            'max_server_per_host=2',
            'affinity_group',
        ]
        verifylist = [
            ('policy', 'soft-anti-affinity'),
            ('rules', {'max_server_per_host': '2'}),
            ('name', 'affinity_group'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.compute_client.create_server_group.assert_called_once_with(
            name=parsed_args.name,
            policy=parsed_args.policy,
            rules=parsed_args.rules,
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_server_group_create_with_rules_pre_v264(self):
        self.set_compute_api_version('2.63')

        arglist = [
            '--policy',
            'soft-anti-affinity',
            '--rule',
            'max_server_per_host=2',
            'affinity_group',
        ]
        verifylist = [
            ('policy', 'soft-anti-affinity'),
            ('rules', {'max_server_per_host': '2'}),
            ('name', 'affinity_group'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-compute-api-version 2.64 or greater is required', str(ex)
        )


class TestServerGroupDelete(TestServerGroup):
    def setUp(self):
        super().setUp()

        self.compute_client.find_server_group.return_value = (
            self.fake_server_group
        )
        self.cmd = server_group.DeleteServerGroup(self.app, None)

    def test_server_group_delete(self):
        arglist = [
            'affinity_group',
        ]
        verifylist = [
            ('server_group', ['affinity_group']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.compute_client.find_server_group.assert_called_once_with(
            'affinity_group', ignore_missing=False
        )
        self.compute_client.delete_server_group.assert_called_once_with(
            self.fake_server_group.id
        )
        self.assertIsNone(result)

    def test_server_group_multiple_delete(self):
        arglist = ['affinity_group', 'anti_affinity_group']
        verifylist = [
            ('server_group', ['affinity_group', 'anti_affinity_group']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.compute_client.find_server_group.assert_any_call(
            'affinity_group', ignore_missing=False
        )
        self.compute_client.find_server_group.assert_any_call(
            'anti_affinity_group', ignore_missing=False
        )
        self.compute_client.delete_server_group.assert_called_with(
            self.fake_server_group.id
        )
        self.assertEqual(2, self.compute_client.find_server_group.call_count)
        self.assertEqual(2, self.compute_client.delete_server_group.call_count)
        self.assertIsNone(result)

    def test_server_group_delete_no_input(self):
        arglist = []
        verifylist = None
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_server_group_multiple_delete_with_exception(self):
        arglist = ['affinity_group', 'anti_affinity_group']
        verifylist = [
            ('server_group', ['affinity_group', 'anti_affinity_group']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.compute_client.find_server_group.side_effect = [
            self.fake_server_group,
            exceptions.CommandError,
        ]
        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 server groups failed to delete.', str(e))

        self.compute_client.find_server_group.assert_any_call(
            'affinity_group', ignore_missing=False
        )
        self.compute_client.find_server_group.assert_any_call(
            'anti_affinity_group', ignore_missing=False
        )
        self.assertEqual(2, self.compute_client.find_server_group.call_count)
        self.compute_client.delete_server_group.assert_called_once_with(
            self.fake_server_group.id
        )


class TestServerGroupList(TestServerGroup):
    def setUp(self):
        super().setUp()

        self.compute_client.server_groups.return_value = [
            self.fake_server_group
        ]
        self.cmd = server_group.ListServerGroup(self.app, None)

    def test_server_group_list(self):
        arglist = []
        verifylist = [
            ('all_projects', False),
            ('long', False),
            ('limit', None),
            ('offset', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.server_groups.assert_called_once_with()

        expected_columns = (
            'ID',
            'Name',
            'Policies',
        )
        expected_data = (
            (
                self.fake_server_group.id,
                self.fake_server_group.name,
                format_columns.ListColumn(self.fake_server_group.policies),
            ),
        )

        self.assertCountEqual(expected_columns, columns)
        self.assertCountEqual(expected_data, tuple(data))

    def test_server_group_list_with_all_projects_and_long(self):
        arglist = [
            '--all-projects',
            '--long',
        ]
        verifylist = [
            ('all_projects', True),
            ('long', True),
            ('limit', None),
            ('offset', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.compute_client.server_groups.assert_called_once_with(
            all_projects=True
        )

        expected_columns = (
            'ID',
            'Name',
            'Policies',
            'Members',
            'Project Id',
            'User Id',
        )
        expected_data = (
            (
                self.fake_server_group.id,
                self.fake_server_group.name,
                format_columns.ListColumn(self.fake_server_group.policies),
                format_columns.ListColumn(self.fake_server_group.member_ids),
                self.fake_server_group.project_id,
                self.fake_server_group.user_id,
            ),
        )

        self.assertCountEqual(expected_columns, columns)
        self.assertCountEqual(expected_data, tuple(data))

    def test_server_group_list_with_limit(self):
        arglist = [
            '--limit',
            '1',
        ]
        verifylist = [
            ('all_projects', False),
            ('long', False),
            ('limit', 1),
            ('offset', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.compute_client.server_groups.assert_called_once_with(limit=1)

    def test_server_group_list_with_offset(self):
        arglist = [
            '--offset',
            '5',
        ]
        verifylist = [
            ('all_projects', False),
            ('long', False),
            ('limit', None),
            ('offset', 5),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.compute_client.server_groups.assert_called_once_with(offset=5)

    def test_server_group_list_v264(self):
        self.set_compute_api_version('2.64')

        arglist = []
        verifylist = [
            ('all_projects', False),
            ('long', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.compute_client.server_groups.assert_called_once_with()

        expected_columns = (
            'ID',
            'Name',
            'Policy',
        )
        expected_data = (
            (
                self.fake_server_group.id,
                self.fake_server_group.name,
                self.fake_server_group.policy,
            ),
        )

        self.assertCountEqual(expected_columns, columns)
        self.assertCountEqual(expected_data, tuple(data))

    def test_server_group_list_with_all_projects_and_long_v264(self):
        self.set_compute_api_version('2.64')

        arglist = [
            '--all-projects',
            '--long',
        ]
        verifylist = [
            ('all_projects', True),
            ('long', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.compute_client.server_groups.assert_called_once_with(
            all_projects=True
        )

        expected_columns = (
            'ID',
            'Name',
            'Policy',
            'Members',
            'Project Id',
            'User Id',
        )
        expected_data = (
            (
                self.fake_server_group.id,
                self.fake_server_group.name,
                self.fake_server_group.policy,
                format_columns.ListColumn(self.fake_server_group.member_ids),
                self.fake_server_group.project_id,
                self.fake_server_group.user_id,
            ),
        )

        self.assertCountEqual(expected_columns, columns)
        self.assertCountEqual(expected_data, tuple(data))


class TestServerGroupShow(TestServerGroup):
    def setUp(self):
        super().setUp()

        self.compute_client.find_server_group.return_value = (
            self.fake_server_group
        )
        self.cmd = server_group.ShowServerGroup(self.app, None)

    def test_server_group_show(self):
        self.set_compute_api_version('2.64')

        arglist = [
            'affinity_group',
        ]
        verifylist = [
            ('server_group', 'affinity_group'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)
